/*
 * Copyright © 2020, 2022 Andreas K. Förster
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * SPDX-License-Identifier: GPL-3.0-or-later
 */

#ifdef __EMSCRIPTEN__
#include <emscripten/emscripten.h>
#endif

#include "SDL.h"
#include <stdlib.h>
#include <time.h>
#include <unistd.h>
#include <limits.h>


#define Programmname "AKF-Mathe"
#define Homepage "https://akfoerster.de/p/akfmathe/"
#define Breite 1024
#define Hoehe  576

#define Sekunden(x)  ((x) * 1000UL)
#define Bedenkzeit  Sekunden (15)

// Zeichenwerte, 0 bis 9 für Ziffern
#define Plus 10
#define Minus 11
#define Mal 12
#define Geteilt 13
#define Gleich 14
#define Zeichenanzahl 15

#define Angriff 0
#define Rueckzug 1

#ifdef __WINDOWS__
#include <windows.h>
#define random()  rand()
#define srandom(x) srand(x)
#define Standardausgabe(s)  Info(s)
#else
#define Standardausgabe(s)  write(STDOUT_FILENO, "" s "", sizeof(s)-1)
#endif

typedef unsigned char Zeichentyp;	// Ziffern und Operatoren (0-14)
typedef unsigned int Zahlentyp;
#define undefiniert UINT_MAX
#define Maximalzahl 999		// passt gerade bei 10*10

enum Zustaende
{ hilfesuchend, auswaehlend, angreifend, erfolgreich, versagt };

enum Sprachtyp
{ Englisch, Deutsch };

struct Aufgabentyp
{
  Uint32 Startzeit;
  Zahlentyp Wert1, Wert2, Ergebnis, Eingabe;
  Zeichentyp Rechenart;
};

struct Audiotyp
{
  SDL_AudioSpec Spezifikation;
  Uint8 *Daten;
  Uint32 Laenge;
};

extern const unsigned char W_Erfolg1[];
extern const unsigned char W_Erfolg2[];
extern const unsigned char W_Erfolg3[];
extern const unsigned char W_Versagen[];
extern const unsigned char B_Icon[];
extern const unsigned char B_Ecke[];
extern const unsigned char B_Zeichen[];
extern const unsigned char B_Brett[];
extern const unsigned char B_Drache[];
extern const unsigned char B_Prinzessin[];
extern const unsigned char B_GPL[];
extern const unsigned char B_Titel_en[];
extern const unsigned char B_Menue_en[];
extern const unsigned char B_Versagen_en[];
extern const unsigned char B_Hilfe_en[];
extern const unsigned char B_Titel_de[];
extern const unsigned char B_Menue_de[];
extern const unsigned char B_Versagen_de[];
extern const unsigned char B_Hilfe_de[];

extern const unsigned int W_Erfolg1_len;
extern const unsigned int W_Erfolg2_len;
extern const unsigned int W_Erfolg3_len;
extern const unsigned int W_Versagen_len;
extern const unsigned int B_Icon_len;
extern const unsigned int B_Ecke_len;
extern const unsigned int B_Zeichen_len;
extern const unsigned int B_Brett_len;
extern const unsigned int B_Drache_len;
extern const unsigned int B_Prinzessin_len;
extern const unsigned int B_GPL_len;
extern const unsigned int B_Titel_en_len;
extern const unsigned int B_Menue_en_len;
extern const unsigned int B_Versagen_en_len;
extern const unsigned int B_Hilfe_en_len;
extern const unsigned int B_Titel_de_len;
extern const unsigned int B_Menue_de_len;
extern const unsigned int B_Versagen_de_len;
extern const unsigned int B_Hilfe_de_len;

static struct Aufgabentyp Aufgabe;

static Uint32 Wartezeit;
static enum Sprachtyp Sprache;
static enum Zustaende Zustand;
static SDL_bool Vollbild, auffrischen, Kiosk;

static SDL_Window *Fenster;
static SDL_Renderer *renderer;

static struct
{
  SDL_AudioDeviceID Audiogeraet;
  struct Audiotyp Erfolg[3], Versagen;
} A;

static struct
{
  SDL_Texture *Titel, *Brett, *Drache, *Prinzessin, *Zeichen, *Logo, *Ecke;
} T;

static struct
{
  SDL_Rect Titel, Brett, Drache, Prinzessin, erleichtert, Zeichen, Logo, Ecke;
  SDL_Rect Boden, Himmel, Pfosten;
} R;


static SDL_NORETURN void Optionshilfe (void);
static SDL_NORETURN void Version (void);
static int initialisieren (void);
static void aufraeumen (void);
static SDL_INLINE void Ende (void);
static void Argumente (int, char *[]);
static SDL_INLINE Zahlentyp Zufallszahl (Zahlentyp, Zahlentyp);
static SDL_INLINE int zentriert (int);
static void Menue (void);
static void Grafikeinrichtung (void);
static void Audioeinrichtung (void);
static SDL_INLINE void Audio (const struct Audiotyp *);
static SDL_Surface *Grafikladen (const unsigned char *, size_t);
static SDL_INLINE void Hintergrundfarbe (void);
static void Ecken (SDL_Renderer *);
static void Schild (SDL_Texture *, SDL_Rect *);
static SDL_NORETURN void Fehler (const char *);
static SDL_INLINE SDL_UNUSED void Info (const char *);
static void Nummerneingabe (const SDL_Keysym *);
static SDL_bool warte (void);
static enum Sprachtyp Systemsprache (void);

static SDL_INLINE void Hauptschleife (void);
static void Ablauf (void);
static void Angriffszene (void);
static void Hilfe (void);
static void Erfolg (void);
static void Versagen (void);

static void Szene (int);	// Szene anzeigen
static void Aufgabenstellung (void);
static void Aufgabenanzeige (void);
static void Zifferneingabe (Zahlentyp);
static SDL_INLINE Uint32 verstrichen (Uint32);
static void Zahl (Zahlentyp, int *, int);
static void Zeichen (Zeichentyp, int *, int);
static SDL_INLINE void Tastaturpuffer_loeschen (void);
void Vollbildumschaltung (void);


int
main (int argc, char *argv[])
{
  Argumente (argc, argv);

  if (initialisieren () < 0)
    Fehler (SDL_GetError ());

  Hauptschleife ();
  return EXIT_SUCCESS;
}


static SDL_INLINE void
Hintergrundfarbe (void)
{
  SDL_SetRenderDrawColor (renderer, 0xF4, 0xEC, 0xD8, SDL_ALPHA_OPAQUE);
}


#ifdef __EMSCRIPTEN__

static SDL_INLINE void
Hauptschleife (void)
{
  emscripten_set_main_loop (Ablauf, 0, 1);
}

#else

static SDL_INLINE void
Hauptschleife (void)
{
  while (1)
    {
      Ablauf ();
      SDL_Delay (16);
    }
}

#endif

static void
Ablauf (void)
{
  switch (Zustand)
    {
    case hilfesuchend:
      Hilfe ();
      break;

    case auswaehlend:
      Menue ();
      break;

    case erfolgreich:
      Erfolg ();
      break;

    case versagt:
      Versagen ();
      break;

    case angreifend:
      Angriffszene ();
      break;
    }
}


static SDL_NORETURN void
Version (void)
{
  switch (Sprache)
    {
    case Deutsch:
      Standardausgabe
	("AKF-Mathe\nCopyright (c) Andreas K. Foerster\n"
	 "Lizenz: GPLv3+ GNU GPL Version 3 oder neuer "
	 "<http://gnu.org/licenses/gpl.html>\n\n"
	 "Dies ist freie Software. Sie darf veraendert und verteilt werden.\n"
	 "Der Quelltext muss verfuegbar gehalten werden.\n"
	 "Es wird keine Garantie gewaehrt, soweit das Gesetz es zulaesst.\n\n"
	 "Homepage: " Homepage "\n");
      break;

    case Englisch:
    default:
      Standardausgabe
	("AKF-Mathe\nCopyright (c) Andreas K. Foerster\n"
	 "License GPLv3+: GNU GPL version 3 or later "
	 "<http://gnu.org/licenses/gpl.html>\n\n"
	 "This is free software; you are free to change and redistribute it.\n"
	 "The source code must be kept available.\n"
	 "There is NO WARRANTY, to the extent permitted by law.\n\n"
	 "Homepage: " Homepage "\n");
      break;
    }

  exit (EXIT_SUCCESS);
}


static SDL_NORETURN void
Optionshilfe (void)
{
  switch (Sprache)
    {
    case Deutsch:
      Standardausgabe
	("AKF-Mathe\n\n"
	 "  -h, --help             Hilfe\n"
	 "  -V, --version          Lizenz\n"
	 "  -f, --fullscreen       Vollbild\n"
	 "  -k                     Kiosk-Modus[!]\n"
	 "  -e                     Englisch\n"
	 "  -d                     Deutsch\n"
	 "  -r n                   Zahlenreihe\n");
      break;

    case Englisch:
    default:
      Standardausgabe
	("AKF-Mathe\n\n"
	 "  -h, --help             Help\n"
	 "  -V, --version          License\n"
	 "  -f, --fullscreen       Fullscreen\n"
	 "  -k                     Kiosk mode[!]\n"
	 "  -e                     English\n"
	 "  -d                     German\n"
	 "  -r n                   number row\n");
      break;
    }

  exit (EXIT_SUCCESS);
}


static void
Argumente (int argc, char *argv[])
{
  Vollbild = Kiosk = SDL_FALSE;
  Sprache = Systemsprache ();
  srandom (time (0));

  if (argc < 2)
    return;

  for (int i = 1; i < argc; ++i)
    {
      char *a = argv[i];

      if (!SDL_strcmp ("--help", a) || !SDL_strcmp ("-h", a))
	Optionshilfe ();
      else if (!SDL_strcmp ("--version", a) || !SDL_strcmp ("-V", a))
	Version ();
      else if (!SDL_strcmp ("-f", a) || !SDL_strcmp ("--fullscreen", a))
	Vollbild = SDL_TRUE;
      else if (!SDL_strcmp ("-k", a))
	Kiosk = Vollbild = SDL_TRUE;
      else if (!SDL_strcmp ("-d", a))
	Sprache = Deutsch;
      else if (!SDL_strcmp ("-e", a))
	Sprache = Englisch;
      else if (!SDL_strcmp ("-r", a) && i < (argc - 1))	// Zahlenreihe
	srandom (SDL_strtoul (argv[++i], NULL, 10));
      else
	Fehler ("Unbekannter Parameter");
    }
}


static SDL_INLINE Zahlentyp
Zufallszahl (Zahlentyp Minimum, Zahlentyp Maximum)
{
  // random () nur hier benutzen!
  return ((Zahlentyp) (random () % (Maximum - Minimum + 1) + Minimum));
}


static void
Aufgabenstellung (void)
{
  Aufgabe.Wert1 = Zufallszahl (1, 10);
  Aufgabe.Wert2 = Zufallszahl (1, 10);
  // nur hierfür Zufallszahlen benutzen! Für feste Zahlenreihen

  switch (Aufgabe.Rechenart)
    {
    case Plus:
      Aufgabe.Ergebnis = Aufgabe.Wert1 + Aufgabe.Wert2;
      break;

    case Minus:
      Aufgabe.Ergebnis = Aufgabe.Wert1;
      Aufgabe.Wert1 += Aufgabe.Wert2;
      break;

    case Mal:
      Aufgabe.Ergebnis = Aufgabe.Wert1 * Aufgabe.Wert2;
      break;

    case Geteilt:
      Aufgabe.Ergebnis = Aufgabe.Wert1;
      Aufgabe.Wert1 *= Aufgabe.Wert2;
      break;

    default:
      Fehler ("Rechenart nicht unterstützt.");
    }

  R.Drache.x = 0 - R.Drache.w;
  Aufgabe.Eingabe = undefiniert;
  Aufgabe.Startzeit = 0;
  Zustand = angreifend;
}


static int
initialisieren (void)
{
  SDL_SetHint (SDL_HINT_VIDEO_ALLOW_SCREENSAVER, Kiosk ? "0" : "1");

  if (SDL_Init (SDL_INIT_VIDEO) < 0)
    return -1;

  atexit (aufraeumen);

  Uint32 Optionen = SDL_WINDOW_RESIZABLE;

  if (Vollbild)
    Optionen |= SDL_WINDOW_FULLSCREEN_DESKTOP;

  Fenster =
    SDL_CreateWindow (Programmname, SDL_WINDOWPOS_CENTERED,
		      SDL_WINDOWPOS_CENTERED, Breite, Hoehe, Optionen);

  if (!Fenster)
    return -1;

  renderer = SDL_CreateRenderer (Fenster, -1, SDL_RENDERER_ACCELERATED);

  if (!renderer)
    return -1;

  SDL_ShowCursor (!Vollbild);
  SDL_SetHint (SDL_HINT_RENDER_SCALE_QUALITY, "linear");
  SDL_RenderSetLogicalSize (renderer, Breite, Hoehe);

  SDL_EventState (SDL_KEYUP, SDL_IGNORE);
  SDL_EventState (SDL_MOUSEMOTION, SDL_IGNORE);
  SDL_EventState (SDL_MOUSEBUTTONDOWN, SDL_IGNORE);
  SDL_EventState (SDL_MOUSEBUTTONUP, SDL_IGNORE);
  SDL_EventState (SDL_MOUSEWHEEL, SDL_IGNORE);

  SDL_ClearError ();

  Hintergrundfarbe ();
  SDL_RenderClear (renderer);
  SDL_RenderPresent (renderer);

  Grafikeinrichtung ();
  Audioeinrichtung ();

  Aufgabe.Startzeit = 0;
  Aufgabe.Rechenart = 0;
  Aufgabe.Eingabe = undefiniert;
  Zustand = auswaehlend;
  Wartezeit = 0;
  auffrischen = SDL_FALSE;

  return 0;
}


// mit atexit aufgerufen
static void
aufraeumen (void)
{
  if (A.Audiogeraet)
    {
      SDL_CloseAudioDevice (A.Audiogeraet);
      SDL_FreeWAV (A.Erfolg[0].Daten);
      SDL_FreeWAV (A.Erfolg[1].Daten);
      SDL_FreeWAV (A.Erfolg[2].Daten);
      SDL_FreeWAV (A.Versagen.Daten);
    }

  SDL_DestroyTexture (T.Logo);
  SDL_DestroyTexture (T.Ecke);
  SDL_DestroyTexture (T.Brett);
  SDL_DestroyTexture (T.Zeichen);
  SDL_DestroyTexture (T.Titel);
  SDL_DestroyTexture (T.Drache);
  SDL_DestroyTexture (T.Prinzessin);

  SDL_DestroyRenderer (renderer);
  SDL_DestroyWindow (Fenster);

  SDL_Quit ();
}


static SDL_Surface *
Grafikladen (const unsigned char *g, size_t l)
{
  SDL_Surface *s;

  s = SDL_LoadBMP_RW (SDL_RWFromConstMem (g, l), SDL_TRUE);

  if (!s)
    Fehler (SDL_GetError ());

  return s;
}


// horizontal zentriert
static SDL_INLINE int
zentriert (int Elementbreite)
{
  return ((Breite / 2) - (Elementbreite / 2));
}


static void
Grafikeinrichtung (void)
{
  SDL_Surface *s;

  s = Grafikladen (B_Icon, B_Icon_len);
  SDL_SetColorKey (s, SDL_TRUE, 15);
  SDL_SetWindowIcon (Fenster, s);
  T.Logo = SDL_CreateTextureFromSurface (renderer, s);

  R.Logo.w = s->w;
  R.Logo.h = s->h;
  R.Logo.y = 10;
  R.Logo.x = Breite - R.Logo.w - R.Logo.y;

  SDL_FreeSurface (s);

  s = Grafikladen (B_Ecke, B_Ecke_len);
  //SDL_SetColorKey (s, SDL_TRUE, 15);
  T.Ecke = SDL_CreateTextureFromSurface (renderer, s);

  R.Ecke.w = s->w;
  R.Ecke.h = s->h;
  R.Ecke.x = R.Ecke.y = 0;

  SDL_FreeSurface (s);

  switch (Sprache)
    {
    case Deutsch:
      s = Grafikladen (B_Titel_de, B_Titel_de_len);
      break;

    case Englisch:
    default:
      s = Grafikladen (B_Titel_en, B_Titel_en_len);
      break;
    }

  //SDL_SetColorKey (s, SDL_TRUE, 15);
  T.Titel = SDL_CreateTextureFromSurface (renderer, s);

  R.Titel.w = s->w;
  R.Titel.h = s->h;
  R.Titel.x = zentriert (R.Titel.w);
  R.Titel.y = 10;

  SDL_FreeSurface (s);

  s = Grafikladen (B_Brett, B_Brett_len);
  SDL_SetColorKey (s, SDL_TRUE, 15);
  T.Brett = SDL_CreateTextureFromSurface (renderer, s);

  R.Brett.w = s->w;
  R.Brett.h = s->h;
  R.Brett.x = zentriert (R.Brett.w);
  R.Brett.y = R.Titel.y + R.Titel.h + 50;

  SDL_FreeSurface (s);

  s = Grafikladen (B_Zeichen, B_Zeichen_len);
  SDL_SetColorKey (s, SDL_TRUE, 15);

  T.Zeichen = SDL_CreateTextureFromSurface (renderer, s);

  if (s->w % Zeichenanzahl)
    Fehler ("Zeichentabelle nicht teilbar");

  R.Zeichen.w = s->w / Zeichenanzahl;
  R.Zeichen.h = s->h;
  R.Zeichen.x = (Breite / 2) - (4 * R.Zeichen.w);
  R.Zeichen.y = R.Brett.y + (R.Brett.h / 2) - (R.Zeichen.h / 2);

  SDL_FreeSurface (s);

  s = Grafikladen (B_Prinzessin, B_Prinzessin_len);
  // 8-Bit-Palette => Pixel links oben transparent
  if (s->format->BytesPerPixel == 1 && s->format->BitsPerPixel == 8)
    SDL_SetColorKey (s, SDL_TRUE, *(Uint8 *) s->pixels);

  T.Prinzessin = SDL_CreateTextureFromSurface (renderer, s);

  R.Prinzessin.w = s->w / 2;
  R.Prinzessin.h = s->h;
  R.Prinzessin.x = Breite - R.Prinzessin.w - 50;
  R.Prinzessin.y = Hoehe - R.Prinzessin.h - 50;

  SDL_FreeSurface (s);

  R.erleichtert.w = R.Prinzessin.w;
  R.erleichtert.h = R.Prinzessin.h;
  R.erleichtert.x = R.Prinzessin.x - 4;
  R.erleichtert.y = R.Prinzessin.y - 4;

  s = Grafikladen (B_Drache, B_Drache_len);
  // 8-Bit-Palette => Pixel links oben transparent
  if (s->format->BytesPerPixel == 1 && s->format->BitsPerPixel == 8)
    SDL_SetColorKey (s, SDL_TRUE, *(Uint8 *) s->pixels);
  T.Drache = SDL_CreateTextureFromSurface (renderer, s);

  R.Drache.w = s->w;
  R.Drache.h = s->h;
  R.Drache.x = 0;
  R.Drache.y = R.Prinzessin.y + R.Prinzessin.h - R.Drache.h - 100;

  SDL_FreeSurface (s);

  R.Himmel.x = R.Himmel.y = R.Boden.x = 0;
  R.Himmel.w = R.Boden.w = Breite;
  R.Himmel.h = R.Boden.y = Hoehe * 3 / 4;
  R.Boden.h = Hoehe - R.Boden.y;

  R.Pfosten.w = 20;
  R.Pfosten.x = zentriert (R.Pfosten.w);
  R.Pfosten.y = R.Brett.y - 10;
  R.Pfosten.h = Hoehe - 100 - R.Pfosten.y;
}


static void
Audioladen (struct Audiotyp *a, const unsigned char *d, size_t l)
{
  SDL_zerop (a);

  if (!SDL_LoadWAV_RW (SDL_RWFromConstMem (d, l), SDL_TRUE,
		       &a->Spezifikation, &a->Daten, &a->Laenge))
    Fehler (SDL_GetError ());
}


static void
Audioeinrichtung (void)
{
  SDL_AudioSpec Spezifikation;

#ifdef __WINDOWS__
  // schneller
  SDL_setenv ("SDL_AUDIODRIVER", "directsound", 0);
#endif

  if (SDL_InitSubSystem (SDL_INIT_AUDIO))
    return;

  SDL_zero (Spezifikation);
  Spezifikation.freq = 24000;
  Spezifikation.format = AUDIO_S16LSB;
  Spezifikation.channels = 1;	// Mono
  Spezifikation.samples = 4096;
  Spezifikation.callback = NULL;
  Spezifikation.userdata = NULL;

  A.Audiogeraet = SDL_OpenAudioDevice (NULL, 0, &Spezifikation, NULL, 0);

  if (A.Audiogeraet)
    {
      Audioladen (&A.Versagen, W_Versagen, W_Versagen_len);
      Audioladen (&A.Erfolg[0], W_Erfolg1, W_Erfolg1_len);
      Audioladen (&A.Erfolg[1], W_Erfolg2, W_Erfolg2_len);
      Audioladen (&A.Erfolg[2], W_Erfolg3, W_Erfolg3_len);
      SDL_PauseAudioDevice (A.Audiogeraet, SDL_FALSE);
    }
}


static SDL_INLINE void
Audio (const struct Audiotyp *a)
{
  if (A.Audiogeraet)
    SDL_QueueAudio (A.Audiogeraet, a->Daten, a->Laenge);
}


static SDL_INLINE Uint32
verstrichen (Uint32 seit)
{
  return (SDL_GetTicks () - seit);
}


static void
Szene (int Modus)
{
  // Hintergrund
  SDL_SetRenderDrawColor (renderer, 0, 0, 0, SDL_ALPHA_OPAQUE);
  SDL_RenderClear (renderer);

  // Boden
  SDL_SetRenderDrawColor (renderer, 0x73, 0xAA, 0x55, SDL_ALPHA_OPAQUE);
  SDL_RenderFillRect (renderer, &R.Boden);

  // Himmel
  SDL_SetRenderDrawColor (renderer, 0xB0, 0xE0, 0xE6, SDL_ALPHA_OPAQUE);
  SDL_RenderFillRect (renderer, &R.Himmel);

  SDL_RenderCopy (renderer, T.Titel, NULL, &R.Titel);

  // Pfosten
  SDL_SetRenderDrawColor (renderer, 0x8B, 0x45, 0x13, SDL_ALPHA_OPAQUE);
  SDL_RenderFillRect (renderer, &R.Pfosten);

  // Schild
  SDL_RenderCopy (renderer, T.Brett, NULL, &R.Brett);

  // Prinzessin
  SDL_Rect Emanation;
  Emanation.w = R.Prinzessin.w;
  Emanation.h = R.Prinzessin.h;
  Emanation.x = Modus ? Emanation.w + 1 : 0;
  Emanation.y = 0;

  SDL_RenderCopy (renderer, T.Prinzessin, &Emanation,
		  Modus ? &R.erleichtert : &R.Prinzessin);

  // Drache
  SDL_RenderCopyEx (renderer, T.Drache, NULL, &R.Drache,
		    0.0, NULL, Modus ? SDL_FLIP_HORIZONTAL : SDL_FLIP_NONE);

  // Logo
  SDL_RenderCopy (renderer, T.Logo, NULL, &R.Logo);
  Aufgabenanzeige ();

  SDL_RenderPresent (renderer);
  auffrischen = SDL_FALSE;
}


static void
Aufgabenanzeige (void)
{
  int x, y;

  x = R.Zeichen.x;
  y = R.Zeichen.y;

  //SDL_RenderSetClipRect (renderer, &R.Brett);

  Zahl (Aufgabe.Wert1, &x, y);
  Zeichen (Aufgabe.Rechenart, &x, y);
  Zahl (Aufgabe.Wert2, &x, y);
  Zeichen (Gleich, &x, y);

  if (Aufgabe.Eingabe != undefiniert)
    Zahl (Aufgabe.Eingabe, &x, y);

  //SDL_RenderSetClipRect (renderer, NULL);
}


static void
Angriffszene (void)
{
  SDL_Event e;
  int alt;
  static int Maul;

  if (!Aufgabe.Startzeit)
    Aufgabe.Startzeit = SDL_GetTicks ();

  alt = Maul;
  Maul = (verstrichen (Aufgabe.Startzeit) * R.Prinzessin.x) / Bedenkzeit;

  if (Maul != alt)
    {
      R.Drache.x = Maul - R.Drache.w;
      Szene (Angriff);

      // Drache fängt die Prinzessin
      if (Maul >= R.Prinzessin.x)
	Zustand = versagt;
    }

  while (SDL_PollEvent (&e))
    switch (e.type)
      {
      case SDL_KEYDOWN:
	Nummerneingabe (&e.key.keysym);
	break;

      case SDL_WINDOWEVENT:
	if (e.window.event == SDL_WINDOWEVENT_EXPOSED)
	  Szene (Angriff);
	break;

      case SDL_QUIT:
	Ende ();
	break;
      }				// switch

  if (Aufgabe.Eingabe == Aufgabe.Ergebnis)
    Zustand = erfolgreich;
}


static void
Zifferneingabe (Zahlentyp Wert)
{
  if (Aufgabe.Eingabe == undefiniert)
    Aufgabe.Eingabe = Wert;
  else
    {
      Zahlentyp n = (Aufgabe.Eingabe * 10) + Wert;
      if (n <= Maximalzahl)
	Aufgabe.Eingabe = n;
    }
}


static void
Nummerneingabe (const SDL_Keysym * k)
{
  switch (k->sym)
    {
    case SDLK_HELP:
    case SDLK_F1:
      Zustand = hilfesuchend;
      break;

    case SDLK_F11:
    case SDLK_f:
      Vollbildumschaltung ();
      break;

    case SDLK_MENU:
    case SDLK_F2:
      Zustand = auswaehlend;
      Aufgabe.Rechenart = 0;
      break;

    case SDLK_ESCAPE:
      Ende ();
      break;

    case SDLK_DELETE:
    case SDLK_BACKSPACE:
    case SDLK_RETURN:
    case SDLK_RETURN2:
    case SDLK_KP_ENTER:
    case SDLK_KP_CLEARENTRY:
    case SDLK_KP_CLEAR:
      Aufgabe.Eingabe = undefiniert;
      Szene (Angriff);
      break;

    case SDLK_0:
    case SDLK_KP_0:
      Zifferneingabe (0);
      break;

    case SDLK_1:
    case SDLK_KP_1:
      Zifferneingabe (1);
      break;

    case SDLK_2:
    case SDLK_KP_2:
      Zifferneingabe (2);
      break;

    case SDLK_3:
    case SDLK_KP_3:
      Zifferneingabe (3);
      break;

    case SDLK_4:
    case SDLK_KP_4:
      Zifferneingabe (4);
      break;

    case SDLK_5:
    case SDLK_KP_5:
      Zifferneingabe (5);
      break;

    case SDLK_6:
    case SDLK_KP_6:
      Zifferneingabe (6);
      break;

    case SDLK_7:
    case SDLK_KP_7:
      Zifferneingabe (7);
      break;

    case SDLK_8:
    case SDLK_KP_8:
      Zifferneingabe (8);
      break;

    case SDLK_9:
    case SDLK_KP_9:
      Zifferneingabe (9);
      break;

    case SDLK_KP_00:
      Aufgabe.Eingabe *= 10;
      Zifferneingabe (0);
      break;

    case SDLK_KP_000:
      Aufgabe.Eingabe *= 100;
      Zifferneingabe (0);
      break;
    }
}


static SDL_bool
warte (void)
{
  SDL_Event e;
  SDL_bool r = SDL_TRUE;

  while (SDL_PollEvent (&e))
    switch (e.type)
      {
      case SDL_KEYDOWN:
	switch (e.key.keysym.sym)
	  {
	  case SDLK_ESCAPE:
	    Ende ();
	    break;

	  case SDLK_F11:
	  case SDLK_f:
	    Vollbildumschaltung ();
	    break;

	  case SDLK_HELP:
	  case SDLK_F1:
	    Zustand = hilfesuchend;
	    Wartezeit = 0;
	    r = SDL_FALSE;
	    break;

	  case SDLK_MENU:
	  case SDLK_F2:
	    Zustand = auswaehlend;
	    Aufgabe.Rechenart = 0;
	    Wartezeit = 0;
	    r = SDL_FALSE;
	    break;

	  default:		// beliebige Taste
	    r = SDL_FALSE;
	    break;
	  }
	break;

      case SDL_WINDOWEVENT:
	if (e.window.event == SDL_WINDOWEVENT_EXPOSED)
	  auffrischen = SDL_TRUE;
	break;

      case SDL_QUIT:
	Ende ();
	break;
      }				// switch

  if (Wartezeit && SDL_TICKS_PASSED (SDL_GetTicks (), Wartezeit))
    r = SDL_FALSE;

  return r;
}


static void
Erfolg (void)
{
  static Uint32 Startzeit;
  static int Startposition;

  if (!Startzeit)
    {
      Zustand = erfolgreich;
      Startzeit = SDL_GetTicks ();
      Wartezeit = Startzeit + Sekunden (4);
      Startposition = R.Drache.x + R.Drache.w;
      Audio (&A.Erfolg[Aufgabe.Wert2 % 3]);
    }

  // Drache zieht sich zurück
  if (R.Drache.x > (-R.Drache.w))
    {
      int x;

      // 2 Sekunden für Rückzug, egal wie weit
      x = Startposition;
      x -= verstrichen (Startzeit) * Startposition / Sekunden (2);
      x -= R.Drache.w;

      if (x != R.Drache.x)
	{
	  R.Drache.x = x;
	  Szene (Rueckzug);
	}

      return;
    }

  if (warte ())
    {
      if (auffrischen)
	Szene (Rueckzug);
    }
  else
    {
      Wartezeit = Startzeit = 0;
      if (Zustand == erfolgreich)
	Aufgabenstellung ();
    }
}


static void
Versagen (void)
{
  static Uint32 start;
  static SDL_Texture *t;
  static SDL_Rect r;
  unsigned int Alpha;

  if (!start)
    {
      SDL_Surface *s;

      Wartezeit = 0;
      start = SDL_GetTicks ();
      Audio (&A.Versagen);

      switch (Sprache)
	{
	case Deutsch:
	  s = Grafikladen (B_Versagen_de, B_Versagen_de_len);
	  break;

	case Englisch:
	default:
	  s = Grafikladen (B_Versagen_en, B_Versagen_en_len);
	  break;
	}

      SDL_SetColorKey (s, SDL_TRUE, 15);

      t = SDL_CreateTextureFromSurface (renderer, s);

      r.w = s->w;
      r.h = s->h;
      r.x = zentriert (r.w);
      r.y = (Hoehe / 2) - (r.h / 2);

      SDL_FreeSurface (s);
    }

  Hintergrundfarbe ();

  Alpha = (verstrichen (start) * SDL_ALPHA_OPAQUE) / Sekunden (2);

  // Einblendung beenden, in Wartemodus übergehen
  if (Alpha < SDL_ALPHA_OPAQUE)
    {
      SDL_SetTextureAlphaMod (t, SDL_min (Alpha, SDL_ALPHA_OPAQUE));
      Schild (t, &r);
      return;
    }
  else if (!Wartezeit)
    {
      Tastaturpuffer_loeschen ();
      SDL_SetTextureAlphaMod (t, SDL_ALPHA_OPAQUE);
      Schild (t, &r);

      Wartezeit = SDL_GetTicks () + Sekunden (30);
      return;
    }

  if (warte ())
    {
      if (auffrischen)
	Schild (t, &r);
    }
  else
    {
      SDL_DestroyTexture (t);
      t = NULL;

      if (Zustand == versagt)
	{
	  if (Wartezeit && SDL_TICKS_PASSED (SDL_GetTicks (), Wartezeit))
	    Zustand = auswaehlend;
	  else
	    Zustand = angreifend;
	}

      Aufgabe.Eingabe = undefiniert;
      Aufgabe.Startzeit = 0;
      start = 0;
      Wartezeit = 0;
    }
}


static void
Hilfeanzeige (void)
{
  SDL_Surface *s;
  SDL_Texture *Text, *GPL;
  SDL_Rect rt, rg;

  switch (Sprache)
    {
    case Deutsch:
      s = Grafikladen (B_Hilfe_de, B_Hilfe_de_len);
      break;

    case Englisch:
    default:
      s = Grafikladen (B_Hilfe_en, B_Hilfe_en_len);
      break;
    }

  SDL_SetColorKey (s, SDL_TRUE, 15);
  Text = SDL_CreateTextureFromSurface (renderer, s);

  rt.w = s->w;
  rt.h = s->h;
  rt.x = zentriert (rt.w);
  rt.y = 50;

  SDL_FreeSurface (s);

  s = Grafikladen (B_GPL, B_GPL_len);
  //SDL_SetColorKey (s, SDL_TRUE, 15);
  GPL = SDL_CreateTextureFromSurface (renderer, s);
  rg.w = s->w;
  rg.h = s->h;
  rg.x = rg.y = 75;
  SDL_FreeSurface (s);

  Hintergrundfarbe ();
  SDL_RenderClear (renderer);
  Ecken (renderer);
  SDL_RenderCopy (renderer, GPL, NULL, &rg);
  SDL_RenderCopy (renderer, Text, NULL, &rt);
  SDL_RenderPresent (renderer);
  auffrischen = SDL_FALSE;

  SDL_DestroyTexture (GPL);
  SDL_DestroyTexture (Text);
}


static void
Hilfe (void)
{
  static SDL_bool angezeigt;
  SDL_Event e;

  if (!angezeigt || auffrischen)
    {
      Hilfeanzeige ();
      angezeigt = SDL_TRUE;
    }

  while (SDL_PollEvent (&e))
    switch (e.type)
      {
      case SDL_KEYDOWN:
	switch (e.key.keysym.sym)
	  {
	  case SDLK_MENU:
	  case SDLK_F2:
	    Zustand = auswaehlend;
	    Aufgabe.Rechenart = 0;
	    break;

	  case SDLK_F11:
	  case SDLK_f:
	    Vollbildumschaltung ();
	    continue;
	    break;

	  case SDLK_ESCAPE:
	    Ende ();
	    break;
	  }

	angezeigt = SDL_FALSE;
	Aufgabe.Eingabe = undefiniert;
	Aufgabe.Startzeit = 0;
	R.Drache.x = 0 - R.Drache.w;
	Zustand = Aufgabe.Rechenart ? angreifend : auswaehlend;
	break;

      case SDL_WINDOWEVENT:
	if (e.window.event == SDL_WINDOWEVENT_EXPOSED)
	  auffrischen = SDL_TRUE;
	break;

      case SDL_QUIT:
	Ende ();
	break;
      }				// switch
}


static void
Ecken (SDL_Renderer * r)
{
  SDL_Rect e;

  e.h = R.Ecke.h;
  e.w = R.Ecke.w;

  e.x = e.y = 0;		// oben links
  SDL_RenderCopyEx (r, T.Ecke, NULL, &e, 0.0, NULL, SDL_FLIP_NONE);
  e.x = Breite - e.w;		// oben rechts
  SDL_RenderCopyEx (r, T.Ecke, NULL, &e, 0.0, NULL, SDL_FLIP_HORIZONTAL);
  e.y = Hoehe - e.h;		// unten rechts
  SDL_RenderCopyEx (r, T.Ecke, NULL, &e, 0.0, NULL,
		    SDL_FLIP_HORIZONTAL | SDL_FLIP_VERTICAL);
  e.x = 0;			// unten links
  SDL_RenderCopyEx (r, T.Ecke, NULL, &e, 0.0, NULL, SDL_FLIP_VERTICAL);

  // Logo
  e.h = R.Logo.h;
  e.w = R.Logo.w;
  e.y = 60;
  e.x = Breite - e.w - e.y;
  SDL_RenderCopy (renderer, T.Logo, NULL, &e);
}


static void
Schild (SDL_Texture * t, SDL_Rect * r)
{
  SDL_RenderClear (renderer);
  Ecken (renderer);
  SDL_RenderCopy (renderer, t, NULL, r);
  SDL_RenderPresent (renderer);
  auffrischen = SDL_FALSE;
}


static void
Menueanzeige (void)
{
  SDL_Surface *s;
  SDL_Texture *t;
  SDL_Rect r;

  switch (Sprache)
    {
    case Deutsch:
      s = Grafikladen (B_Menue_de, B_Menue_de_len);
      break;

    case Englisch:
    default:
      s = Grafikladen (B_Menue_en, B_Menue_en_len);
      break;
    }

  //SDL_SetColorKey (s, SDL_TRUE, 15);

  if (Kiosk)
    s->h -= 74;			// Vollbild und Ende nicht anzeigen

  t = SDL_CreateTextureFromSurface (renderer, s);

  r.w = s->w;
  r.h = s->h;
  r.x = zentriert (r.w);
  r.y = (Hoehe / 2) - (r.h / 2);

  SDL_FreeSurface (s);

  Hintergrundfarbe ();
  SDL_RenderClear (renderer);
  Ecken (renderer);
  SDL_RenderCopy (renderer, t, NULL, &r);
  SDL_DestroyTexture (t);

  SDL_RenderPresent (renderer);
  auffrischen = SDL_FALSE;
}


static void
Menue (void)
{
  static SDL_bool angezeigt;
  SDL_Event e;

  Zustand = auswaehlend;
  Aufgabe.Rechenart = 0;

  if (!angezeigt || auffrischen)
    {
      Menueanzeige ();
      angezeigt = SDL_TRUE;
    }

  while (SDL_PollEvent (&e))
    switch (e.type)
      {
      case SDL_KEYDOWN:
	switch (e.key.keysym.sym)
	  {
	  case SDLK_HELP:
	  case SDLK_F1:
	    angezeigt = SDL_FALSE;
	    Zustand = hilfesuchend;
	    break;

	  case SDLK_ESCAPE:
	    Ende ();
	    break;

	  case SDLK_F11:
	  case SDLK_f:
	    Vollbildumschaltung ();
	    break;

	  case SDLK_PLUS:
	    // deutsche Tastatur
	    Aufgabe.Rechenart = (e.key.keysym.mod & KMOD_SHIFT) ? Mal : Plus;
	    break;

	  case SDLK_KP_PLUS:
	    Aufgabe.Rechenart = Plus;
	    break;

	  case SDLK_MINUS:
	  case SDLK_KP_MINUS:
	    Aufgabe.Rechenart = Minus;
	    break;

	  case SDLK_ASTERISK:
	  case SDLK_KP_MULTIPLY:
	    Aufgabe.Rechenart = Mal;
	    break;

	  case SDLK_SLASH:
	  case SDLK_COLON:
	  case SDLK_KP_DIVIDE:
	    Aufgabe.Rechenart = Geteilt;
	    break;

	  case SDLK_7:
	  case SDLK_PERIOD:
	    // deutsche Tastatur
	    if (e.key.keysym.mod & KMOD_SHIFT)
	      Aufgabe.Rechenart = Geteilt;
	    break;
	  }
	break;

      case SDL_WINDOWEVENT:
	if (e.window.event == SDL_WINDOWEVENT_EXPOSED)
	  auffrischen = SDL_TRUE;
	break;

      case SDL_QUIT:
	Ende ();
	break;
      }				// switch

  if (Aufgabe.Rechenart)
    {
      angezeigt = SDL_FALSE;
      Aufgabenstellung ();
    }
}


static SDL_INLINE void
Tastaturpuffer_loeschen (void)
{
  SDL_PumpEvents ();
  SDL_FlushEvent (SDL_KEYDOWN);
}


// gibt Englisch oder Deutsch aus - nichts anderes
static enum Sprachtyp
Systemsprache (void)
{
  char *l;
  enum Sprachtyp e;

  e = Englisch;

  // Posix
  if ((l = SDL_getenv ("LC_ALL"))
      || (l = SDL_getenv ("LC_MESSAGES")) || (l = SDL_getenv ("LANG")))
    {
      if (SDL_strncasecmp (l, "de", 2) == 0)
	e = Deutsch;
    }

#ifdef __WINDOWS__
  // auch unter Windows die Umgebungsvariablen respektieren
  else if (PRIMARYLANGID (GetUserDefaultLangID ()) == LANG_GERMAN)
    e = Deutsch;
#endif

  return e;
}


// z zwischen 0 und 14 (abgesichert)
static void
Zeichen (Zeichentyp z, int *x, int y)
{
  SDL_Rect Bild, Ziel;

  Bild.w = Ziel.w = R.Zeichen.w;
  Bild.h = Ziel.h = R.Zeichen.h;

  Bild.x = (z % Zeichenanzahl) * Bild.w;
  Bild.y = 0;

  Ziel.x = *x;
  Ziel.y = y;

  SDL_RenderCopy (renderer, T.Zeichen, &Bild, &Ziel);

  *x += Ziel.w;
}


// gibt vorzeichenlose Ganzzahl aus
static void
Zahl (Zahlentyp Wert, int *x, int y)
{
  if (Wert < 10)
    Zeichen ((Zeichentyp) Wert, x, y);
  else
    {
      Zeichentyp *p, Puffer[21];

      // Wert von hinten in den Puffer schreiben
      // der Puffer ist nötig, um die Reihenfolge umzudrehen
      p = Puffer + sizeof (Puffer) - 1;
      *p = Gleich;		// markiert Ende

      do
	{
	  --p;
	  *p = (Zeichentyp) (Wert % 10);	// letzte Ziffer
	  Wert /= 10;
	}
      while (Wert != 0 && p > &Puffer[0]);

      // Ausgabe
      while (*p < 10)
	Zeichen (*p++, x, y);
    }
}


static SDL_NORETURN void
Fehler (const char *Meldung)
{
  SDL_ShowSimpleMessageBox (SDL_MESSAGEBOX_ERROR, Programmname, Meldung,
			    Fenster);
  exit (EXIT_FAILURE);
}


// derzeit nur bei Windows verwendet
static SDL_INLINE SDL_UNUSED void
Info (const char *Meldung)
{
  SDL_ShowSimpleMessageBox (SDL_MESSAGEBOX_INFORMATION, Programmname,
			    Meldung, Fenster);
}


static SDL_INLINE void
Ende (void)
{
  if (!Kiosk)
    exit (EXIT_SUCCESS);
}


// extern fuer wasm
void
Vollbildumschaltung (void)
{
  if (Kiosk)
    return;

  Vollbild = !Vollbild;
  SDL_SetWindowFullscreen (Fenster,
			   Vollbild ? SDL_WINDOW_FULLSCREEN_DESKTOP : 0);
  SDL_ShowCursor (!Vollbild);
  auffrischen = SDL_TRUE;
}
